/*
 * Copyright (c) 2012, Kustaa Nyholm / SpareTimeLabs
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list 
 * of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice, this 
 * list of conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 *  
 * Neither the name of the Kustaa Nyholm or SpareTimeLabs nor the names of its 
 * contributors may be used to endorse or promote products derived from this software 
 * without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 */
package purejavacomm;

import purejavacomm.*;

import java.io.*;
import java.nio.ByteBuffer;

import com.sun.jna.Native;
import com.sun.jna.NativeLong;
import com.sun.jna.Platform;

import jtermios.*;
import jtermios.linux.JTermiosImpl.Linux_C_lib;
import jtermios.linux.JTermiosImpl.Linux_C_lib.pollfd;
import jtermios.linux.JTermiosImpl.Linux_C_lib.serial_struct;
import jtermios.linux.JTermiosImpl.Linux_C_lib.termios;
import jtermios.linux.JTermiosImpl.Linux_C_lib.timeval;
import static jtermios.JTermios.*;

public class RawStream {
	static public InputStream getInputStream(SerialPort port) throws IOException {
		if (!(port instanceof PureJavaSerialPort))
			return port.getInputStream();
		if (Platform.isWindows())
			return port.getInputStream();
		int FD = ((PureJavaSerialPort) port).getNativeFileDescriptor();
		if (NativeLong.SIZE == 4)
			return new RawPosix32bInputStream(FD);
		if (NativeLong.SIZE == 8)
			return new RawPosix64bInputStream(FD);
		return null;
	}

	static public void configureThresholdTimeout(SerialPort port, int threshold, int timeout) throws UnsupportedCommOperationException, IOException {
		if (!(port instanceof PureJavaSerialPort) || Platform.isWindows()) {
			port.enableReceiveTimeout(timeout);
			port.enableReceiveThreshold(threshold);
			return;
		}
		if (timeout < 0 || timeout > 25550 || threshold < 0 || threshold > 255)
			throw new UnsupportedCommOperationException();
		int FD = ((PureJavaSerialPort) port).getNativeFileDescriptor();
		Termios termios = new Termios();
		if (0 != tcgetattr(FD, termios))
			throw new IOException();

		termios.c_cc[VTIME] = (byte) ((timeout + 99) / 100);
		termios.c_cc[VMIN] = (byte) threshold;

		if (0 != tcsetattr(FD, TCSANOW, termios))
			throw new IOException();

	}

	static class RawPosixInputStream extends InputStream {
		int m_FD;
		private int[] m_Available = { 0 };
		private CLib m_CLib = new CLib();

		protected RawPosixInputStream(int FD) {
			m_FD = FD;
		}

		public static class CLib {
			native public int ioctl(int fd, int cmd, int[] arg);

			static {
				try {
					Native.register("c");
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}

		@Override
		final public int available() throws IOException {
			if (m_CLib.ioctl(m_FD, FIONREAD, m_Available) < 0)
				throw new IOException();
			return m_Available[0];
		}

		@Override
		final public int read() throws IOException {
			byte[] buf = { 0 };
			int n = read(buf, 0, 1);
			return n > 0 ? buf[0] & 0xFF : -1;
		}

	}

	static class RawPosix32bInputStream extends RawPosixInputStream {
		CLib m_CLib = new CLib();

		public static class CLib {
			native public int read(int fd, byte[] buffer, int count);

			native public int read(int fd, ByteBuffer buffer, int count);

			static {
				try {
					Native.register("c");
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}

		public RawPosix32bInputStream(int fd) {
			super(fd);
		}

		@Override
		final public int read(byte[] buffer) throws IOException {
			return m_CLib.read(m_FD, buffer, buffer.length);
		}

		@Override
		final public int read(byte[] buffer, int offset, int length) throws IOException {
			if (offset == 0)
				return m_CLib.read(m_FD, buffer, length);
			else
				return m_CLib.read(m_FD, ByteBuffer.wrap(buffer, offset, length), length);
		}

	}

	static class RawPosix64bInputStream extends RawPosixInputStream {
		CLib m_CLib = new CLib();

		public static class CLib {
			native public long read(int fd, byte[] buffer, long length);

			native public long read(int fd, ByteBuffer buffer, long length);

			static {
				try {
					Native.register("c");
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}

		public RawPosix64bInputStream(int fd) {
			super(fd);
		}

		@Override
		final public int read(byte[] buffer) throws IOException {
			return (int) m_CLib.read(m_FD, buffer, (long) buffer.length);
		}

		@Override
		final public int read(byte[] buffer, int offset, int length) throws IOException {
			if (offset == 0)
				return (int) m_CLib.read(m_FD, buffer, (long) length);
			else
				return (int) m_CLib.read(m_FD, ByteBuffer.wrap(buffer, offset, length), (long) length);
		};

	}

}
